/* criteria.h -*- C++ -*- criteria interface
 * Rob Earhart
 * $Id: criteria.h,v 1.6 1998/03/09 00:58:05 rob Exp $
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#ifndef ACAP_CRITERIA_H
#define ACAP_CRITERIA_H

#include "ptr.h"
#include "str.h"
#include "compare.h"

class EntryRef;

// Criteria are unary functions which return true if the Entry does
// *not* match the criteria (i.e. if the entry in question should be
// skipped).  Yeah, that's a little weird; I'm defining it this way so
// that they can be easily used with remove_if (composing them with
// logical_not<bool> would involve extra work because compose1 needs
// to copy its unary_function arguments).

// ... although not as easily as I'd though, since remove_if also
// copies its unary_function argument.  Suck.  So, we use CriteriaPtr, below...

// Yes, I'm aware that the singular of criteria is criterion, and
// that the singular criteria, although increasingly common in both
// speech and printed media over the last fifty years, is still
// considered non-standard.  But that's okay; much like the technology
// I develop, my vocabulary is strictly on the cutting edge (thanks to
// Dan Martinez <url:mailto:dmartine@adobe.com> for bringing my
// attention to this issue).  -- )Rob

class Criteria : public unary_function<EntryRef *, bool> {
public:
  virtual ~Criteria() {}
  virtual bool operator()(EntryRef *entry) = 0;
};

class CritALL : public Criteria {
public:
  ~CritALL();
  bool operator()(EntryRef *);
};

class CritAND : public Criteria {
public:
  CritAND(Ptr<Criteria> &_first,
	  Ptr<Criteria> &_second)
    : first(_first), second(_second) {}
  ~CritAND();
  bool operator()(EntryRef *entry);
private:
  Ptr<Criteria> first, second;
};

class CritCOMPARE : public Criteria {
public:
  CritCOMPARE(const String &_attribute,
	      const String &comparator,
	      const String &_value)
    :  ordering(Compare::Instance()->LookupOrdering(comparator)),
       attribute(_attribute),
       value(_value) {}
  ~CritCOMPARE();
  bool operator()(EntryRef *entry);
private:
  Compare::ordering_t ordering;
  String attribute, value;
};

class CritCOMPARESTRICT : public Criteria {
public:
  CritCOMPARESTRICT(const String &_attribute,
		    const String &comparator,
		    const String &_value)
    : ordering(Compare::Instance()->LookupOrdering(comparator)),
      attribute(_attribute),
      value(_value) {}
  ~CritCOMPARESTRICT();
  bool operator()(EntryRef *entry);
private:
  Compare::ordering_t ordering;
  String attribute, value;
};

class CritEQUAL : public Criteria {
public:
  CritEQUAL(const String &_attribute,
	    const String &comparator,
	    Ptr<String> _value)
    : equality(Compare::Instance()->LookupEquality(comparator)),
      attribute(_attribute),
      value(_value) {}
  ~CritEQUAL();
  bool operator()(EntryRef *entry);
private:
  Compare::equality_t equality;
  String attribute;
  Ptr<String> value;
};

class CritNOT : public Criteria {
public:
  CritNOT(Ptr<Criteria> &_clause)
    : clause(_clause) {}
  ~CritNOT();
  bool operator()(EntryRef *entry);
private:
  Ptr<Criteria> clause;
};

class CritOR : public Criteria {
public:
  CritOR(Ptr<Criteria> &_first,
	 Ptr<Criteria> &_second)
    : first(_first), second(_second) {}
  ~CritOR();
  bool operator()(EntryRef *entry);
private:
  Ptr<Criteria> first, second;
};

class CritPREFIX : public Criteria {
public:
  CritPREFIX(const String &_attribute,
	     const String &comparator,
	     const String &_value)
    : match(Compare::Instance()->LookupMatch(comparator)),
      attribute(_attribute),
      value(_value) {}
  ~CritPREFIX();
  bool operator()(EntryRef *entry);
private:
  Compare::match_t match;
  String attribute, value;
};

class CritRANGE : public Criteria {
public:
  CritRANGE(unsigned long low,
	    unsigned long high,
	    const String &value);
  ~CritRANGE();
  bool operator()(EntryRef *entry);
private:
  unsigned long low, high;
  String value;
};

class CritSUBSTRING : public Criteria {
public:
  CritSUBSTRING(const String &_attribute,
		const String &comparator,
		const String &_value)
    : match(Compare::Instance()->LookupMatch(comparator)),
      attribute(_attribute),
      value(_value) {}
  ~CritSUBSTRING();
  bool operator()(EntryRef *entry);
private:
  Compare::match_t match;
  String attribute, value;
};


// Wrapper to get around function object copying weirdness
class CriteriaPtr : public unary_function<EntryRef *, bool> {
public:
  CriteriaPtr(Criteria *_c) : c(_c) {}
  bool operator()(EntryRef *entry) { return (*c)(entry); }
private:
  Criteria *c;
};

#endif // ACAP_CRITERIA_H
