/* dsobjects.h -*- C++ -*- dataset persistent object definitions
 * Rob Earhart
 * $Id: dsobjects.h,v 1.2 1998/03/09 00:58:14 rob Exp $
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#ifndef ACAP_DSOBJECTS_H
#define ACAP_DSOBJECTS_H

#include "ps.h"
#include "trx.h"

typedef enum {
  empty_tag = 0,		// necessary for newly allocated pages,
  // which *should* be zero-filled...
  dataset_tag,
  entrylist_tag,
  string_tag,
  attrlist_tag,
  stringlist_tag,
  max_tag
  // Note: try to keep bits/tag < bits/char, eh?
  // Or change the definition of an object_t in ps.h.
} object_tag;

// Notes:
//
// Everything that is intended to be pointed to (foo_t) should start
// with an object_t.
//
// Everything that is intended to be used as a component (fooval_t)
// should *not* have a constructor; values will be set in-place.

// Note: Ordinarily, I'd make list_t point to an array of buckets;
// that's the Right Way to do it, right?  Well... not if you're
// swizzling; the container and the contained objects must be an
// single object.  So this is a little (heh) ugly.
template<class O, object_tag tag>
struct list_t {
  void *operator new(size_t size, PS *ps, unsigned n_buckets,
		     Trx *trx, off_t *offset) {
    PS::alloc_t a = ps->Alloc(size + n_buckets * sizeof(O), trx);
    if (offset) *offset = a.second;
    return a.first;
  }
  list_t(unsigned _n_buckets, Trx *trx)
    : base((char)tag, trx), next(0), n_buckets(_n_buckets), n_alloced(0) {
      trx->WriteOffset(PageOffsetOfAddr(&next), 0);
      trx->WriteData(PageOffsetOfAddr(&n_buckets),
		     &n_buckets,
		     sizeof(n_buckets));
      trx->WriteData(PageOffsetOfAddr(&n_alloced),
		     &n_alloced,
		     sizeof(n_alloced));
  }
  void Swizzle(PS *ps) {
    base.Swizzle(ps);
    SwizzlePtr(ps, next);
    for (unsigned lupe = 0; lupe < n_alloced; lupe++) {
      DEBUG("Swizzling bucket " << lupe);
      buckets(lupe)->Swizzle(ps);
    }
  }
  O *alloc_bucket(Trx *trx) {
    assert(n_alloced < n_buckets);
    n_alloced++;
    construct(buckets(n_alloced - 1), trx);
    trx->WriteData(PageOffsetOfAddr(&n_alloced),
		   &n_alloced,
		   sizeof(n_alloced));
    return buckets(n_alloced - 1);
  }
  void schedule_free(PS *ps, PSBlob *depender, Trx *trx) {
    for (unsigned lupe = 0; lupe < n_alloced; lupe++)
      (*this)[lupe].schedule_free(ps, depender, trx);
    if (next) next->schedule_free(ps, depender, trx);
    depender->AddDependant(ps->ScheduleFree(this, trx));
  }

  O &operator[](unsigned ind) {
    assert(ind < n_alloced);
    return ((O *)(this + 1))[ind];
  }
  O *buckets(unsigned ind) {
    assert(ind < n_alloced);
    return ((O *)(this + 1)) + ind;
  }

  PS::object_t base;
  list_t *next;		// placeholder for large lists
  unsigned n_buckets;		// amount of space we have for buckets
  unsigned n_alloced;		// number of buckets actually in use
};

struct string_t {
  typedef String::iterator iterator;
  typedef String::const_iterator const_iterator;
  void *operator new(size_t size, PS *ps,
		     String::size_type len,
		     Trx *trx, off_t *offset) {
    PS::alloc_t a = ps->Alloc(size + len, trx);
    if (offset) *offset = a.second;
    return a.first;
  }
  string_t(const String &str, Trx *trx) : base(string_tag, trx) {
    next = NULL;
    trx->WriteOffset(PageOffsetOfAddr(&next), 0);
    size = str.size();
    trx->WriteData(PageOffsetOfAddr(&size),
		   &size,
		   sizeof(size));
    memcpy(value, str.octets(), size);
    trx->WriteData(PageOffsetOfAddr(&value),
		   &value,
		   size);
  }
  void Swizzle(PS *ps) {
    base.Swizzle(ps);
  }

  iterator begin() { return value; }
  const_iterator begin() const { return value; }
  iterator end() { return value + size; }
  const_iterator end() const { return value + size; }

  void schedule_free(PS *ps, PSBlob *depender, Trx *trx) {
    depender->AddDependant(ps->ScheduleFree(this, trx));
  }

  PS::object_t base;
  string_t *next;
  String::size_type size;
  String::value_type value[1];
};

struct stringval_t {
  stringval_t(Trx *trx) : hash(0), str(NULL) {
    trx->WriteData(PageOffsetOfAddr(this),
		   this,
		   sizeof(*this));
  }
  stringval_t() : hash(0), str(NULL) {
    // If we're called without arguments, we're embedded as part
    // of some other structure, which will write us out as
    // it initializes.  (The other case is that we're a bucket
    // in a stringlist_t - but list_t uses construct() to setup
    // its buckets, which passes a trx to the constructor).
  }
  void Swizzle(PS *ps) {
    SwizzlePtr(ps, str);
  }
  void assign(const stringval_t &it, PS *ps, Trx *trx) {
    hash = it.hash;
    trx->WriteData(PageOffsetOfAddr(&hash),
		   &hash,
		   sizeof(hash));
    str = it.str;
    trx->WriteOffset(PageOffsetOfAddr(&str), ps->OffsetOfAddress(str));
  }
  bool operator==(const String &it) const {
    return str
      && hash == it.hash()
      && str->size == it.size()
      && ! lexicographical_compare_3way(str->begin(), str->end(),
					it.begin(), it.end());
  }
  void set(PS *ps, const String &it, Trx *trx) {
    assert(! str);
    hash = it.hash();
    trx->WriteData(PageOffsetOfAddr(&hash),
		   &hash,
		   sizeof(hash));
    off_t offset;
    str = new(ps, it.size(), trx, &offset) string_t(it, trx);
    trx->WriteOffset(PageOffsetOfAddr(&str), offset);
  }
  void schedule_free(PS *ps, PSBlob *depender, Trx *trx) {
    if (str) str->schedule_free(ps, depender, trx);
  }

  String::hash_type hash;
  string_t *str;
};

typedef list_t<stringval_t, stringlist_tag> stringlist_t;

struct attrval_t {
  attrval_t(Trx *trx)
    : createtime(trx->TrxStamp()),
      deletetime(Timestamp::MaxStamp()),
      havectype(false),
      values(NULL),
      acls(NULL) {
	trx->WriteData(PageOffsetOfAddr(this),
		       this,
		       sizeof(*this));
  }
  void Swizzle(PS *ps) {
    name.Swizzle(ps);
    SwizzlePtr(ps, values);
    contenttype.Swizzle(ps);
    SwizzlePtr(ps, acls);
  }
  void set_name(PS *ps, const String &it, Trx *trx) {
    name.set(ps, it, trx);
  }
  void assign(const attrval_t &it, PS *ps, Trx *trx) {
    name.assign(it.name, ps, trx);
    createtime = it.createtime;
    trx->WriteData(PageOffsetOfAddr(&createtime),
		   &createtime,
		   sizeof(createtime));
    deletetime = it.deletetime;
    trx->WriteData(PageOffsetOfAddr(&deletetime),
		   &deletetime,
		   sizeof(deletetime));
    state = it.state;
    trx->WriteData(PageOffsetOfAddr(&state),
		   &state,
		   sizeof(state));
    havectype = it.havectype;
    trx->WriteData(PageOffsetOfAddr(&havectype),
		   &havectype,
		   sizeof(havectype));
    values = it.values;
    trx->WriteOffset(PageOffsetOfAddr(&values), ps->OffsetOfAddress(values));
    contenttype = it.contenttype;
    trx->WriteData(PageOffsetOfAddr(&contenttype),
		   &contenttype,
		   sizeof(contenttype));
    acls = it.acls;
    trx->WriteOffset(PageOffsetOfAddr(&acls), ps->OffsetOfAddress(acls));
  }
  bool in_date(Timestamp::Stamp stamp) const {
    return createtime <= stamp
      && stamp < deletetime;
  }
  void set_delete(Timestamp::Stamp stamp,
		  Trx *trx) {
    deletetime = stamp;
    trx->WriteData(PageOffsetOfAddr(&deletetime),
		   &deletetime,
		   sizeof(deletetime));
  }
  void set_havectype(Trx *trx) {
    havectype = true;
    trx->WriteData(PageOffsetOfAddr(&contenttype),
		   &contenttype,
		   sizeof(contenttype));
  }
  void set_state_nil(Trx *trx) {
    state = nil;
    trx->WriteData(PageOffsetOfAddr(&state),
		   &state,
		   sizeof(state));
  }
  void set_state_def(Trx *trx) {
    state = def;
    trx->WriteData(PageOffsetOfAddr(&state),
		   &state,
		   sizeof(state));
  }
  void set_state_single(Trx *trx) {
    state = single;
    trx->WriteData(PageOffsetOfAddr(&state),
		   &state,
		   sizeof(state));
  }
  void set_state_multi(Trx *trx) {
    state = multi;
    trx->WriteData(PageOffsetOfAddr(&state),
		   &state,
		   sizeof(state));
  }
  void set_value(stringlist_t *sl, off_t offset, Trx *trx) {
    assert(! values);
    values = sl;
    trx->WriteOffset(PageOffsetOfAddr(&values), offset);
  }

  void schedule_free(PS *ps, PSBlob *depender, Trx *trx) {
    name.schedule_free(ps, depender, trx);
    if (values) values->schedule_free(ps, depender, trx);
    contenttype.schedule_free(ps, depender, trx);
    if (acls) acls->schedule_free(ps, depender, trx);
  }

  stringval_t name;
  Timestamp::Stamp createtime;
  Timestamp::Stamp deletetime;
  typedef enum {
    nil, def, single, multi
  } state_t;
  state_t state;
  bool havectype;
  stringlist_t *values;
  stringval_t contenttype;
  stringlist_t *acls;
};

typedef list_t<attrval_t, attrlist_tag> attrlist_t;

struct entryval_t {
  entryval_t(Trx *trx)
    : createtime(trx->TrxStamp()),
      modtime(trx->TrxStamp()),
      deletetime(Timestamp::MaxStamp()),
      inherit(true),
      freeattrs(true),
      attrs(NULL) {
	trx->WriteData(PageOffsetOfAddr(this),
		       this,
		       sizeof(*this));
  }
  void Swizzle(PS *ps) {
    name.Swizzle(ps);
    SwizzlePtr(ps, attrs);
  }
  void set_name(PS *ps, const String &it, Trx *trx) {
    name.set(ps, it, trx);
  }
  void assign(const entryval_t &it, PS *ps, Trx *trx) {
    name.assign(it.name, ps, trx);
    set_createtime(it.createtime, trx);
    set_modtime(it.modtime, trx);
    set_deletetime(it.deletetime, trx);
    set_inherit(it.inherit, trx);
    set_freeattrs(it.freeattrs, trx);
    set_attrs(it.attrs, trx, ps);
  }
  bool in_date(Timestamp::Stamp stamp) const {
    return createtime <= stamp
      && stamp < deletetime;
  }
  void set_createtime(Timestamp::Stamp stamp, Trx *trx) {
    createtime = stamp;
    trx->WriteData(PageOffsetOfAddr(&createtime),
		   (void *)&createtime,
		   sizeof(createtime));
  }
  void set_modtime(Timestamp::Stamp stamp, Trx *trx) {
    modtime = stamp;
    trx->WriteData(PageOffsetOfAddr(&modtime),
		   (void *)&modtime,
		   sizeof(modtime));
  }
  void set_deletetime(Timestamp::Stamp stamp, Trx *trx) {
    deletetime = stamp;
    trx->WriteData(PageOffsetOfAddr(&deletetime),
		   (void *)&deletetime,
		   sizeof(deletetime));
  }
  void schedule_free(PS *ps, PSBlob *depender, bool do_attrs, Trx *trx) {
    name.schedule_free(ps, depender, trx);
    if (do_attrs && freeattrs && attrs) {
      DEBUG("XXX SCHEDULING ATTR FREE (attrs at " << attrs << ")");
      attrs->schedule_free(ps, depender, trx);
    }
  }
  void set_attrs(attrlist_t *as, Trx *trx, off_t offset) {
    DEBUG("Setting attrs to " << as << " (offset " << (void *)offset << ")");
    attrs = as;
    trx->WriteOffset(PageOffsetOfAddr(&attrs), offset);
  }
  void set_attrs(attrlist_t *as, Trx *trx, PS *ps) {
    DEBUG("Setting attrs to " << as);
    attrs = as;
    trx->WriteOffset(PageOffsetOfAddr(&attrs), ps->OffsetOfAddress(attrs));
  }
  void set_inherit(bool _inherit, Trx *trx) {
    inherit = _inherit;
    trx->WriteData(PageOffsetOfAddr(&inherit),
		   (void *)&inherit,
		   sizeof(inherit));
  }
  void set_freeattrs(bool _freeattrs, Trx *trx) {
    freeattrs = _freeattrs;
    trx->WriteData(PageOffsetOfAddr(&freeattrs),
		   (void *)&freeattrs,
		   sizeof(freeattrs));
  }
  stringval_t name;
  Timestamp::Stamp createtime;
  Timestamp::Stamp modtime;
  Timestamp::Stamp deletetime;
  bool inherit;
  bool freeattrs;
  attrlist_t *attrs;
};

typedef list_t<entryval_t, entrylist_tag> entrylist_t;

struct datasetobject_t {
  void *operator new(size_t size, PS *ps, Trx *trx, off_t *offset) {
    PS::alloc_t a = ps->Alloc(size, trx);
    if (offset) *offset = a.second;
    return a.first;
  }
  datasetobject_t(PS *ps, Trx *trx)
    : base(dataset_tag, trx), entries(NULL), indicies(NULL) {
      trx->WriteOffset(PageOffsetOfAddr(&entries), 0);
      trx->WriteOffset(PageOffsetOfAddr(&indicies), 0);
  }
  void Swizzle(PS *ps) {
    base.Swizzle(ps);
    SwizzlePtr(ps, entries);
  }
  void set_entries(entrylist_t *ents, Trx *trx, off_t offset) {
    entries = ents;
    trx->WriteOffset(PageOffsetOfAddr(&entries), offset);
  }
  PS::object_t base;
  Timestamp::Stamp deletedvalidity;
  entrylist_t *entries;
  void *indicies;		// unused
};

#endif // ACAP_DSOBJECTS_H
