/* interlock.cc - dataset interlock implementation
 * Rob Earhart
 * $Id: interlock.cc,v 1.5 1998/03/09 00:58:27 rob Exp $
 */
/***********************************************************
        Copyright 1997 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#include "interlock.h"

Interlock *Interlock::instance = NULL;
pthread_once_t Interlock::instance_init = PTHREAD_ONCE_INIT;

void Interlock::Initialize() {
  instance = new Interlock();
}

Interlock *Interlock::Instance() {
  if (! instance)
    pthread_once(&instance_init, &Initialize);
  return instance;
}

Interlock::Resource::Resource(const String &_name) : name(_name) {
  DEBUG("Created resource " << Name());
}

Interlock::Resource::~Resource() {
  DEBUG("Removing resource " << Name());
  Interlock::Instance()->RemoveResource(name);
}

void Interlock::Resource::Noref() {
  Grab g(Interlock::Instance()->InterlockMutex());
  if (CheckNoRefs())
    delete this;
}

void Interlock::Resource::Wait(Timestamp::Stamp stamp, locks_t &l) {
  Grab g(locks_mutex);

  // Here, we just loop around, testing to see whether we're clear
  // with the given list (i.e. we're lower than any timestamp in the
  // list), and if not, wait for the highest thing to complete (since
  // it'll probably the last to complete, although if not, no big
  // deal, that's why we loop).  We could probably do away with the
  // loop by using a semaphore which'd block us until everything had
  // completed, but this is easier, and should work just as
  // well... (heh - posix semaphores aren't implemented under the
  // version of Solaris we're using anyway).
  while(1) {
    locks_t::iterator it = l.lower_bound(stamp);
    if (it == l.begin()) break;
    --it;
    (*it).second->Wait(locks_mutex);
  }
}

void Interlock::Resource::Wait(Timestamp::Stamp stamp, bool exists) {
  // Okay.  First, we clear with the locks list, and then if it
  // doesn't exist, we clear with the implicit list.
  Wait(stamp, locks);
  if (! exists)
    Wait(stamp, implicit);
}

void Interlock::Resource::RemoveVersion(Version *v) {
  DEBUG("Removing version from resource " << Name());
  Grab g(locks_mutex);
  locks_t::iterator it = locks.find(v->GetTimestamp()->GetStamp());
  if (it != locks.end()) {
    (*it).second->Broadcast();
    locks.erase(it);
  }
  it = implicit.find(v->GetTimestamp()->GetStamp());
  if (it != implicit.end()) {
    (*it).second->Broadcast();
    implicit.erase(it);
  }
}

void Interlock::Resource::AddVersion(Version *v, bool implicit_lock) {
  DEBUG("Adding version on resource " << Name());
  Grab g(locks_mutex);
  locks_t::value_type value(v->GetTimestamp()->GetStamp(), new Cond());
  if (implicit_lock)
    implicit.insert(value);
  else
    locks.insert(value);
}

Interlock::Version::Version()
  : timestamp(TimestampFactory::Instance()->NewStamp()) {
}

Interlock::Version::~Version() {
  // This removes this version from all of the resources it's
  // associated with.
  Grab g(Interlock::Instance()->InterlockMutex());
  for (resources_t::iterator it = resources.begin();
       it != resources.end();
       ++it)
    (*it)->RemoveVersion(this);
}

void Interlock::Version::AddResource(const String &name, bool implicit_create) {
  P<Resource> resource = Interlock::Instance()->LookupResource(name);
  resources.insert(resource);	// if already there, so much the better...
  resource->AddVersion(this, implicit_create);
}

void Interlock::WaitForResource(const String &name,
				Timestamp::Stamp stamp,
				bool exists) {
  // This little dance is here to make sure that the resource doesn't
  // go away at a bad time...
  P<Resource> resource;
  {
    Grab g(InterlockMutex());
    resources_t::iterator it = resources.find(name);
    if (it == resources.end())
      return;
    resource = (*it).second;
  }
  resource->Wait(stamp, exists);
}

P<Interlock::Resource> Interlock::LookupResource(const String &name) {
  DEBUG("Looking up resource " << name);
  resources_t::iterator it = resources.find(name);
  if (it == resources.end())
    it = (resources.insert(resources_t::value_type(name, new Resource(name)))).first;
  return (*it).second;
}

void Interlock::RemoveResource(const String &name) {
  resources.erase(name);
}
