/* notifycontext.cc - notifyable context implementation
 * Rob Earhart
 * $Id: notifycontext.cc,v 1.8 1998/03/09 00:58:37 rob Exp $
 */
/***********************************************************
        Copyright 1997 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#include <hash_map.h>
#include "void.h"
#include "notifycontext.h"
#include "conn.h"
#include "clientstr.h"
#include "sort.h"
#include "dataset.h"
#include "dsgroup.h"
#include "entryref.h"

// NotifyPaths is a singleton class which is responsible for keeping track
// of the path/context mappings
class NotifyPaths {
public:
  static NotifyPaths *Instance() {
    if (! instance)
      pthread_once(&instance_init, &Initialize);
    return instance;
  }
  void NotifyDatasetChange(String path);
  void NotifyPathChange(String path);
  void UnlinkContext(String path, NotifyContext *context); // requires pathsmutex to be held
  void LinkContextDataset(String path, NotifyContext *context);
  void LinkContextPath(String path, NotifyContext *context);
  Mutex &PathsMutex() { return pathsmutex; }
private:
  static void Initialize() {
    instance = new NotifyPaths();
  }
  static NotifyPaths *instance;
  static pthread_once_t instance_init;

  typedef hash_set<NotifyContext *, hash<void *>,
    		   equal_to<void *> > contexts_t;
  // first in pair is path contexts,
  // second is dataset contexts
  typedef hash_map<String, pair<contexts_t, contexts_t>,
  		   hash<String>, equal_to<String> > pathmap_t;
  pathmap_t pathmap;
  Mutex pathsmutex;
};

void NotifyPaths::NotifyDatasetChange(String path) {
  Grab g(PathsMutex());
  pathmap_t::const_iterator it = pathmap.find(path);
  if (it == pathmap.end()) return;
  for (contexts_t::const_iterator cit = (*it).second.second.begin();
       cit != (*it).second.second.end();
       ++cit)
    (*cit)->ScheduleUpdate();	// note: this *may* be a partial update, at some time. XXX
}

void NotifyPaths::NotifyPathChange(String path) {
  Grab g(PathsMutex());
  pathmap_t::const_iterator it = pathmap.find(path);
  if (it == pathmap.end()) return;
  for (contexts_t::const_iterator cit = (*it).second.first.begin();
       cit != (*it).second.first.end();
       ++cit)
    (*cit)->ScheduleUpdate();	// note: this *must* be a full update.
}

void NotifyPaths::UnlinkContext(String path, NotifyContext *context) {
  pathmap_t::iterator it = pathmap.find(path);
  if (it == pathmap.end()) return;
  (*it).second.first.erase(context);
  (*it).second.second.erase(context);
  if ((*it).second.first.empty()
      && (*it).second.second.empty())
    pathmap.erase(it);
}

void NotifyPaths::LinkContextPath(String path, NotifyContext *context) {
  Grab g(PathsMutex());
  pathmap[path].first.insert(context);
}

void NotifyPaths::LinkContextDataset(String path, NotifyContext *context) {
  Grab g(PathsMutex());
  pathmap[path].second.insert(context);
}

NotifyPaths *NotifyPaths::instance = NULL;
pthread_once_t NotifyPaths::instance_init = PTHREAD_ONCE_INIT;

NotifyContext::NotifyContext(const String &_name,
			     Conn *_conn,
			     const Timestamp::Stamp &_modtime,
			     const String &_path,
			     bool _enumerated)
  : Context(_conn, _modtime, _enumerated),
    name(_name),
    path(_path),
    needs_update(false), has_update_task(true) {
      DEBUG("Creating context " << Name());
      if (! IsDatasetPath(path)) {
	Grab g(Client()->PeerContextsMutex());
	Client()->AddPeer(path, this);
      }
}

NotifyContext::~NotifyContext() {
  DEBUG("Destroying context " << Name());
}

void NotifyContext::Noref() {
  Grab g;
  if (! IsDatasetPath(path)) {
    g.Acquire(Client()->PeerContextsMutex());
  } else {
    g.Acquire(NotifyPaths::Instance()->PathsMutex());
  }
  if (! CheckNoRefs())
    return;
  if (! IsDatasetPath(path)) {
    Client()->RemovePeer(path, this);
  } else {
    DropDependancies();
  }
  delete this;
}

void NotifyContext::DropDependancies() {
  // Note: requires that the NotifyPaths::PathsMutex() be held,
  // and that nothing else access this context.
  for (paths_t::const_iterator it = paths.begin();
       it != paths.end();
       ++it)
    NotifyPaths::Instance()->UnlinkContext(*it, this);
  {
    paths_t p;
    paths.swap(p);
  }
}

class Updater : public Task {
public:
  Updater(NotifyContext *_context) : context(_context) {
    context->has_update_task = true;
  }
  void Run() throw () {
    while(1) {
      DEBUG("Context updater: looping");
      {
	Grab g(context->flag_mutex);
	context->needs_update = false;
      }
      try {
	Ptr<Timestamp> ts = Interlock::Instance()->CreateStamp();
	context->Update(ts);
      } catch (...) {}
      {
	Grab g(context->flag_mutex);
	if (! context->needs_update) {
	  context->has_update_task = false;
	  break;
	}
      }
    }
  }
private:
  P<NotifyContext> context;
};

void NotifyContext::ScheduleUpdate() {
  DEBUG("Scheduling update of context " << Name());
  Grab g(flag_mutex);
  needs_update = true;
  if (! has_update_task) {
    Ptr<Task> task = new Updater(this);
    Dispatch::Instance()->Enqueue(task);
  }
}

static const String slash("/");

void NotifyContext::IssueChange(unsigned old_pos,
				unsigned new_pos,
				const EntryRef *entry) const {
  DEBUG("Update: Sending a Change");
  if (! Enumerated())
    old_pos = new_pos = 0;
  Grab g(Client()->OutputMutex());
  Client()->Out() << "* Change " << ClientStr(Name()) << ' '
		  << ClientStr(entry->RelPath())
		  << ' ' << old_pos << ' ' << new_pos;
  if (modifiers->GotReturn()) {
    entry->Write(Client()->Out(),
		 modifiers->ReturnList().begin(),
		 modifiers->ReturnList().end());
  }
  Client()->Out() << "\r\n";
}

void NotifyContext::IssueAddTo(unsigned pos,
			       const EntryRef *entry) const {
  DEBUG("Update: Sending an AddTo");
  if (! Enumerated())
    pos = 0;
  Grab g(Client()->OutputMutex());
  Client()->Out() << "* AddTo " << ClientStr(Name()) << ' '
		  << ClientStr(entry->RelPath())
		  << ' ' << pos;
  if (modifiers->GotReturn()) {
    entry->Write(Client()->Out(),
		 modifiers->ReturnList().begin(),
		 modifiers->ReturnList().end());
  }
  Client()->Out() << "\r\n";
}

void NotifyContext::IssueRemoveFrom(unsigned pos,
				    const EntryRef *entry) const {
  DEBUG("Update: Sending a RemoveFrom");
  if (! Enumerated())
    pos = 0;
  Grab g(Client()->OutputMutex());
  Client()->Out() << "* RemoveFrom " << ClientStr(Name()) << ' '
		  << ClientStr(entry->RelPath())
		  << ' ' << pos << "\r\n";
}

void NotifyContext::Update(Timestamp *ts) {
  assert(ts);

  Grab g(update_mutex);
  DEBUG("Beginning update of context " << Name());

  bool is_modified = false;
  bool issued_change = false;
  P<Searchable> searchable;

  DEBUG("Update: Looking up " << path);

  {
    Grab d(NotifyPaths::Instance()->PathsMutex());
    DropDependancies();
  }

  if (IsDatasetPath(path))
    try {
      searchable = Dataset::TraversePath(path, ts->GetStamp(), NULL, this);
    } catch (Response::NoExist) {}
  else {
    Grab g(Client()->ContextsMutex());
    Conn::contexts_t::iterator it = Client()->Contexts().find(path);
    if (it != Client()->Contexts().end())
      searchable = (*it).second;
  }

  DEBUG("Update: Got current dataset for " << path);

  entries_t new_entries;

  Ptr<DatasetGroup> newdsgroup = new DatasetGroup();

  if (searchable) {
    DEBUG("Update: Performing full search");
    searchable->Search(new_entries, newdsgroup, this, path, modifiers, criteria, NULL, ts->GetStamp());
  }

  DEBUG("Update: Search complete");

  // Okay.  Now we do a diff from entries to new_entries.
  // This is a little tricky, but generates a minimal diff
  // in linear time, so hey.  The macro, in particular, is
  // a little gross (but it's better to have all the code
  // in one place than repeated four times with
  // slightly different variations.)

  // The found_t hash maps defined below keep track of where we've
  // seen something.

  struct hash_found {
    size_t operator()(const EntryRef *ent) const {
      return ent->RelPath().hash();
    }
  };
  struct eq_found {
    size_t operator()(const EntryRef *a, const EntryRef *b) const {
      return a->RelPath() == b->RelPath();
    }
  };
  typedef hash_map<EntryRef *, unsigned, hash_found, eq_found> found_t;
  
  found_t found_in_old, found_in_new;

  // ... and here's the gross macro.  It checks the found_t map for
  // the list that it didn't find it in; if it's not there, we save it
  // in the found_t map for the list we found it in, in case we find
  // it later in the other list, and if it's there, we remove it and
  // issue a CHANGE.

#define NOTE(found_l, found_iter, found_nl, found_counter, func) {	\
    found_t::iterator it = found_nl.find(*found_iter);			\
    if (it == found_nl.end()) {						\
      /* not found - save location */					\
      DEBUG("Not found - saving location");				\
      found_l[*found_iter] = found_counter;				\
    } else {								\
      /* found - we can issue a CHANGE */				\
      DEBUG("Found - issuing CHANGE");					\
      issued_change = true;						\
      func;								\
      found_nl.erase(it);						\
    }									\
    /* ... and get on with life */					\
    ++found_iter;							\
    found_counter++;							\
  }

  dsgroup->Open();		// we need to be able to access our entries

  entries_t::const_iterator o = entries.begin(),
      			    n = new_entries.begin();
  unsigned pos = 1, new_pos = 1;
  Sort comp(modifiers->SortList());
  while (o != entries.end() && n != new_entries.end()) {
    DEBUG("Looping: pos is " << pos << "; new_pos is " << new_pos);
    // DEBUG("Modtime is " << modtime);
    // DEBUG("(*n)->ModTime() is " << (*n)->ModTime());
    // DEBUG("*n is " << *n << "; *o is " << *o);
    DEBUG("Checking old: " << (*o)->RelPath() << " against modtime and new: " << (*n)->RelPath());
    if (modtime < (*n)->ModTime() || (**o) != (**n)) {
      is_modified = true;	// there's been *some* modification;
				// we'll need to inform our peers.
      // They're different; we need to figure out which side
      // is less, deal with it, and increment its iterator
      DEBUG("Comparing " << (*o)->RelPath() << " against " << (*n)->RelPath());
      int c = comp.compare(*o, *n);
      DEBUG("Compare returned " << c);
      if (c < 0) {
	// original item is smaller - if we've seen it (in found_in_new),
	// we issue a CHANGE ('cause hey, here it is!); otherwise, we
	// cache it away in found_in_original.
	NOTE(found_in_old, o, found_in_new, pos,
	     IssueChange(pos, (*it).second, (*it).first));
      } else if (0 < c) {
	// new item is smaller
	NOTE(found_in_new, n, found_in_old, new_pos,
	     IssueChange((*it).second, new_pos, (*it).first));
      } else {
	// they compare equal, but we're here, so *something*
	// happened, so we issue an in-place CHANGE and increment both
	// iterators and both positions.
	DEBUG("Issuing in-place change");
	IssueChange(pos, new_pos, *n);
	issued_change = true;
	++o;
	++n;
	pos++, new_pos++;
      }
    } else {
      // They're identical and haven't changed, so we go on to the next
      // items on both sides.
      DEBUG("No change - continuing");
      ++o;
      ++n;
      pos++, new_pos++;
    }
  }

  // Now that we're through with one side or the other, we can
  // go through what's remaining, and issue CHANGEs
  DEBUG("Issuing remaining old changes");
  while(o != entries.end()) {
    NOTE(found_in_old, o, found_in_new, pos,
	 IssueChange(pos, (*it).second, (*it).first));
  }
  DEBUG("Issuing remaining new changes");
  while(n != new_entries.end()) {
    NOTE(found_in_new, n, found_in_old, new_pos,
	 IssueChange((*it).second, new_pos, (*it).first));
  }
  
  // And then we can send ADDTOs and REMOVEFROMs as appropriate
  // for the things found in one list but not the other.
  DEBUG("Issuing ADDTOs");
  found_t::const_iterator it;
  for (it = found_in_old.begin();
       it != found_in_old.end();
       ++it) {
    issued_change = true;
    is_modified = true;
    IssueRemoveFrom((*it).second, (*it).first);
  }
  DEBUG("Issuing REMOVEFROMs");
  for (it = found_in_new.begin();
       it != found_in_new.end();
       ++it) {
    issued_change = true;
    is_modified = true;
    IssueAddTo((*it).second, (*it).first);
  }

  {
    // Save new entries and dataset group
    Grab g(entries_mutex);
    entries.swap(new_entries);
    dsgroup = newdsgroup;
    dsgroup->Close();
  }

  DEBUG("Update: completing notification");

  // Issue a modtime, if needed
  if (issued_change) {
    Grab g(Client()->OutputMutex());
    Client()->Out() << "* Modtime " << ClientStr(Name())
		    << " \"" << ts->LongStr() << "\"\r\n";
    Client()->Out().flush();
  }
  modtime = ts->GetStamp();

  // And let people know
  if (is_modified) {
    DEBUG("Update: signalling peers");
    Grab g(Client()->PeerContextsMutex());
    Client()->NotifyPeers(Name());
  }
  DEBUG("Update complete");
}

void NotifyContext::DoneInitializing(NotifyContext::entries_t &ents,
				     Ptr<DatasetGroup> &group,
				     Ptr<Criteria> &crit,
				     Ptr<Modifiers> &mods) {
  Context::DoneInitializing(ents, group, crit, mods);
  criteria = crit;
  modifiers = mods;
  Grab g(flag_mutex);
  has_update_task = false;
  if (needs_update) {
    Ptr<Task> task = new Updater(this);
    Dispatch::Instance()->Enqueue(task);
  }
}

void NotifyContext::Enumerate(NotifyContext::entries_t &result,
			      DatasetGroup *group,
			      Timestamp::Stamp stamp) {
  {
    Grab g(StateMutex()); // wait for this context to be ready
  }
  Grab g(entries_mutex);	// make sure no one's playing with the entries
  result.reserve(entries.size());
  for (entries_t::const_iterator it = entries.begin();
       it != entries.end();
       ++it) {
    Ptr<EntryRef> ent = new EntryRef(**it);
    result.push_back(ent);
  }
  group->AddGroup(dsgroup);
}

void NotifyContext::AddPathDependancy(String p) {
  DEBUG("Context " << Name() << " watching path " << p);
  NotifyPaths::Instance()->LinkContextPath(p, this);
  paths.insert(p);
}

void NotifyContext::AddDatasetDependancy(String p) {
  DEBUG("Context " << Name() << " watching dataset " << p);
  NotifyPaths::Instance()->LinkContextDataset(p, this);
  paths.insert(p);
}

void NotifyContext::NotifyDatasetChange(String path) {
  DEBUG("Notifying contexts watching dataset " << path);
  NotifyPaths::Instance()->NotifyDatasetChange(path);
}

void NotifyContext::NotifyPathChange(String path) {
  DEBUG("Notifying contexts watching path " << path);
  NotifyPaths::Instance()->NotifyPathChange(path);
}

