/* page.cc - page implementation
 * Rob Earhart
 * $Id: page.cc,v 1.2 1998/03/09 00:58:39 rob Exp $
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#include <signal.h>
#include "page.h"
#include "ps.h"
#include "unixerror.h"
#include "init.h"

Pages *Pages::instance = NULL;
pthread_once_t Pages::instance_init = PTHREAD_ONCE_INIT;

Pages::Pages() : tmpsize(0) {
  tmp = tmpfile();
  if (! tmp) throw UnixError("Creating file for temporary data");
}

void Pages::Initialize() {
  instance = new Pages();
}

void Pages::FaultHandler(int sig, siginfo_t *sip, void *up) {
  caddr_t addr = sip->_data._fault._addr;
  caddr_t base = (caddr_t)((unsigned)addr & ~pagesize_m1);
  assert(base <= addr && addr < base + pagesize);
  off_t swap_offset;
  P<Page> page;
  //DEBUG("Faulted on address " << (void *)addr << ", base of " << (void *)base);
  {
    Grab g(MapMutex());
    mappings_t::iterator it = mappings.find(base);
    if (it == mappings.end()) {
      // Uh Oh.  We don't have a mapping for this address, so we
      // really can't fault it in from anything.  This is almost
      // certainly indicative of a serious bug in the program;
      // unfortunately, the only action we can really take is to
      // abort(), generate a core, and do postmortem debugging.
      // (Maybe in the future, we should write BYE notifications to
      // all open connections when we abort(), using onexit handlers?
      // Hrm... XXX)
      abort();
    }
    page = (*it).second;
    if (! page) return;		// while we were faulting, another thread
				// completely swizzled the page and removed
				// the Page object; we can go back to work.
  } // drop the map mutex so we're not holding it while we wait for
    // the page to become ready.

  // We need to find out if the page has already been swizzled in.
  // We keep this grab while we're swizzling it, to force every other
  // thread waiting for this page to wait until we're done...
  Grab g(page->mutex);

  if (page->swizzled) return;	// another thread did this already - we
				// can go on now...

  // Otherwise, we need to swizzle.  So we start by grabbing some
  // mmap()-addressable memory from our tmp file...
  {
    Grab g(MapMutex());
    // alloc a page
    if (! freeswap.empty()) {
      swap_offset = freeswap.front();
      freeswap.pop_front();
    } else {
      // otherwise, we need to allocate some space...
      swap_offset = tmpsize;
      tmpsize += pagesize;
      if (ftruncate(fileno(tmp), tmpsize)) abort();
    }
  } // end of block in which we hold the map_mutex

  // map it in
  caddr_t tempbase = mmap(NULL,
			  pagesize,
			  PROT_READ | PROT_WRITE,
			  MAP_SHARED,
			  fileno(tmp), swap_offset);
  if (tempbase == MAP_FAILED) {
    DEBUG("Map failed - aborting (errno == " << errno << ")");
    abort();
  }

  // XXX Deal with systems that don't have pread (Ugh - that'd mean
  // throwing a lock, effectively killing the number of pages which
  // could be simultaneously read for a given PS...)
  PS *ps = page->ps;
  if (pread(ps->GetFD(), tempbase, pagesize, page->offset) != pagesize)
    abort();

  // swizzle it
  ps->SwizzlePage(base, tempbase);

  // unmap the original mapping - we do this first so that we'll have
  // a better chance of dealing with machines that use reverse page
  // tables, which can't deal with having the same page of data mapped
  // to two different locations in the global virtual address space
  // (however uncommon that may be - hpux 9 does it, anything else?).
  munmap(tempbase, pagesize);

  // remap it so that it appears on top of the original page, with the
  // correct protections (this effectively gets us our atomic
  // swizzle&unprotect operation, which we need in order to do
  // swizzling in a threaded environment).
  {
    Grab g(ps->PSMutex());	// make *sure* no one's playing with this PS's pages...
    caddr_t newmap = mmap(base, pagesize, PROT_READ | PROT_WRITE,
			  MAP_PRIVATE | MAP_FIXED, fileno(tmp), swap_offset);
    assert(newmap == base);
  }
  
  // Write to it... (volatile to force the compiler to not optimize
  // this out).  This causes us to actually gain our own private copy
  // of the page in question, so we can change the underlying file.
  *(volatile off_t *)base = *(volatile off_t *)base;

  // And cleanup
  page->swizzled = true;
  {
    Grab g(MapMutex());

    // Give back the page of freeswap...
    freeswap.push_front(swap_offset);

    // And dump the now-unneeded page object.  We can't just erase the
    // address->page mapping, although I'd love to; we have no way to
    // discover whether or not some other thread has faulted on this
    // page and is about to enter the PageFaultHandler and look up the
    // address; if the address is missing, it'll think that it faulted
    // on a bad address, and it'll promptly abort() the program.
    // Since a thread could theoretically take a completely arbitrary
    // amount of time between faulting and getting to the handler, we
    // can Never remove the mapping itself, until we know that no
    // thread could possibly in that state, which is why we do it only
    // when the PS itself goes away (faulting threads will keep a
    // reference to the PS).

    mappings_t::iterator it = mappings.find(base);
    (*it).second = NULL;
  }

  // Falling out here lets go of our reference to the Page, so only
  // threads that came in while we were working should still have it.
}

void Pages::PageFaultHandler(int sig, siginfo_t *sip, void *up) {
  Pages::Instance()->FaultHandler(sig, sip, up);
}

void SetupPFHandler() {
  struct sigaction sa;
  sa.sa_handler = NULL;
  sa.sa_sigaction = &Pages::PageFaultHandler;
  sigemptyset(&sa.sa_mask);
  sa.sa_flags = SA_RESTART | SA_SIGINFO;
  if (sigaction(SIGSEGV, &sa, NULL))
    throw UnixError("Catching SIGSEGV");
}

static Init init(&SetupPFHandler);
