/* ps.h -*- C++ -*- persistent store interface
 * Rob Earhart
 * $Id: ps.h,v 1.2 1998/03/09 00:58:43 rob Exp $
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#ifndef ACAP_PS_H
#define ACAP_PS_H

// For some comments on how all this works, check out page.h.

#include <hash_map.h>
#include <multimap.h>
#include <slist.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <signal.h>
#include "util.h"
#include "page.h"
#include "refcount.h"
#include "void.h"

class PS;
class Trx;

// I'd prefer to make these PS::SwizzlePtr and PS::PageOffsetOfAddr,
// but some c++ compilers don't deal with member templates...
template<class T>
inline void SwizzlePtr(PS *ps, T *&obj) {
  //DEBUG("Swizzling off_t " << (void *)obj);
  obj = (T *)ps->AddressOfOffset((off_t)obj);
  //DEBUG("Swizzled it to address " << (void *)obj);
}

// This calculates the file offset of any address,
// assuming that the address lies in a properly initialized page of
// file memory (i.e. one which starts with the corresponding file offset).
// It's intended to be a little faster than looking up the address->offset
// mapping in a hash table, like OffsetOfAddress does.

template<class T>
inline off_t PageOffsetOfAddr(T *addr) {
  return ((off_t)(addr) & pagesize_m1 | *((off_t *)((off_t)(addr) & ~pagesize_m1)));
}

// Note: We assume in quite a few places that sizeof(off_t) == sizeof(void *);
// in practice, this is definitely *not* always true (n32 on IRIX), but it's
// true enough for most systems.  Still, it should be fixed at some point...
// Maybe instead of off_t, we should just use (unsigned long)
// everywhere?  But that wouldn't work on IRIX64.  We really should
// just use union { off_t, void * } and forget about it... (a template
// for our pointer types would solve the whole mess very nicely).

class PSBlob : public Refcounted {
public:
  // Note: only PS should create blobs.
  PSBlob(PS *_ps, void *_addr);
  void ScheduleFree();
  ~PSBlob();
  void Noref();
  // This whole "dependancy" stuff makes me a little uncomfortable;
  // it's *great*, in one sense, because it means that you only
  // have to hold onto the top level of some tree of data which
  // you'd prefer to not have deleted out from under you (which avoids
  // a lot of swizzling); nevertheless, it *really* makes me wish
  // for a decent PS GC; if a circular loop is accidently created (shouldn't
  // ever happen, but...), data will never be freed, and it'll be next
  // to impossible to catch the problem.
  void AddDependant(PSBlob *blob) {
    dependants.push_front(blob);
  }
private:
  P<PS> ps;
  typedef slist<P<PSBlob> > dependants_t;
  dependants_t dependants;
  void *addr;
  bool free_scheduled;
};

// A resources object exists to keep track of object modifications.
// Each transaction requests a resources object from each persistent
// store it modifies, and lets go of it when the data's been flushed
// all the way back down to the underlying persistent store file.
// When the resources object is no longer referenced, it lets the PS
// know that all of its data has been flushed, so the PS can dump
// all its in-core copies of all its pages, considerably reducing
// memory useage.
class PSResources : public Refcounted {
public:
  ~PSResources();
  void Noref();
private:
  friend class PS;
  PSResources(PS *ps);
  P<PS> ps;
};
  
// Refcounted is virtual so that both Searchable and PS can inherit from it,
// and Dataset can inherit from both.
class PS : virtual public Refcounted {
public:
  PS(const String &_path, int _fd = -1)
    : inited(false), path(_path), fd(_fd), root(NULL), resources(NULL) {}
  void Init(Trx *trx);
  ~PS();

  P<PSResources> GetResources();

  // Allocated a chunk of persistent memory
  typedef pair<caddr_t, off_t> alloc_t;
  alloc_t Alloc(size_t nbytes, Trx *trx);

  // AddressOfOffset and OffsetOfAddress assume that the PSMutex() is
  // held; they perform translations between addresses and offsets.
  caddr_t AddressOfOffset(off_t offset);
  off_t OffsetOfAddress(void *ptr);

  // ScheduleFree writes the free transaction,
  // and sets things up to actually reuse the memory
  // after all pending transactions are finished
  // and after the memory's been reclaimed.
  P<PSBlob> ScheduleFree(void *addr, Trx *trx);

  // SwizzlePage is used by the Pages singleton class to ask a PS to
  // swizzle in a page of data.  Realaddr's where the page will
  // eventually live; addr's the place it lives now.
  void SwizzlePage(caddr_t realaddr, caddr_t addr);

  // Blob interface functions
  // Things which need to keep blobs of data around need to request
  // blobs (using GetBlob) for those bits of memory.  When the blobs
  // go away, the memory may be recycled.
  P<PSBlob> GetBlob(void *addr);

  // Blob->PS interfaces
  Mutex &PSMutex() { return ps_mutex; }
  void RemoveBlob(void *addr) {
    // used by PSBlob::~PSBlob(); ps_mutex must be held
    blobs.erase(addr);
  }
  // used only by PSBlob, hopefully - recycles associated memory,
  // and needs the ps_mutex to be held.
  void Recycle(void *addr);

  // used by the Pages singleton for reading in the underlying data
  // for a given page.
  int GetFD() { return fd; }

  caddr_t Root() const { return root; }

  // FlushPages should be called when the persistent store's data probably
  // won't be referenced for a little while; it doesn't deallocate any address
  // space, but it does deallocate the associated physical memory (swizzling
  // it back in later if needed).
  void FlushPages();

  struct object_t {
    object_t(char tag, Trx *trx);
    void Swizzle(PS *ps) {}	// nothing to swizzle; this function's
				// just here for completeness


    // I'm using a struct here 'cause I'd prefer it if the compiler
    // would smoosh these together into a single word, so that on
    // machines where an off_t is a word we can align to words, so
    // we'll only have a single word of overhead per page.
    struct {
      unsigned tag : CHAR_BIT;
      unsigned size : pagebits;
    } header;
  };

  const String &Path() { return path; }
protected:
  virtual caddr_t CreateRoot(Trx *trx, off_t *offset) = 0;
  virtual void SwizzleObject(caddr_t object, unsigned char tag) = 0;

private:
  typedef union {
    char object[sizeof(object_t)];
    off_t offset;
  } align_t;

  typedef enum {
    map_file,
    map_zero
  } map_type;

  // Note: MapPage assumes that the ps_mutex is held...
  caddr_t MapPage(off_t offset, map_type, int prot);

  // Used for offset->address mapping for a given PS
  typedef hash_map<off_t, caddr_t,
                   PageOffsetHash, equal_to<off_t> > addresses_t;
  // ... and address->offset mapping
  typedef hash_map<caddr_t, off_t,
                   PageOffsetHash, equal_to<caddr_t> > offsets_t;
  // ... and for free memory management
  typedef multimap<size_t, off_t, less<size_t> > freemem_t;

  // Data
  bool inited;
  addresses_t addresses;
  offsets_t offsets;
  freemem_t freemem;
  Mutex ps_mutex;
  String path;
  FD fd;
  struct stat sb;

  caddr_t root;

  typedef hash_map<void *, PSBlob *, hash<void *>, equal_to<void *> > blobs_t;
  blobs_t blobs;

  // Used to keep track of resources
  friend class PSResources;
  Mutex resources_mutex;
  PSResources *resources;
};

#endif // ACAP_PS_H
