/* refcount.h -*- C++ -*- refcounted pointer interface
 * Rob Earhart
 * $Id: refcount.h,v 1.3 1998/02/18 21:19:15 rob Exp $
 */
/***********************************************************
        Copyright 1997 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#ifndef ACAP_REFCOUNT_H
#define ACAP_REFCOUNT_H

// WARNING: This is a fairly evil and twisted thing to do.
// Refcount garbage collection is quite possible the worst
// there is - it's slow, and doesn't deal with cycles in data.
// EXPLICIT GC SHOULD BE USED WHERE POSSIBLE - this is only
// intended as a last resort (and should probably be replaced with
// a real garbage collector at some point).
// This is particularly evil since we're in a threaded environment,
// and we don't have atomic increment and decrememnt instructions,
// making every ref/deref *expensive*...

#include <assert.h>
// This picks up bool
#include <stl_config.h>
#include "debug.h"
#include "grab.h"
#include "mutex.h"

// We dance around a little when unreferencing an object, for the
// convenience of classes which declare their own Noref routine (in
// case the object can be referred to in some other way, say, through
// a table of cached objects), in which case, we need to unlock the
// object, let the Noref routine need lock the table, *then* relock
// and check the object to see if it's acquired a reference in the
// meantime (using CheckNoRefs()).

// The default Noref function will do the right thing (since it
// assumes there's no way to get a reference to an object with
// no known references).

class Refcounted {
public:
  Refcounted() : refcount(0) {}
  virtual ~Refcounted() {}
  void Ref() {
    Grab g(ref_mutex);
    refcount++;
  }
  void Unref() {
    ref_mutex.Lock();
    assert(refcount);
    if (refcount == 1) {
      ref_mutex.Unlock();
      Noref();
    } else {
      refcount--;
      ref_mutex.Unlock();
    }
  }

protected:
  // Replacement Noref functions MUST ensure that nothing can
  // reference the object (by locking all possible places where
  // something could get a reference), and then MUST call
  // CheckNoRefs, which will grab the object, unreference it, and
  // return true if the object now has no references.
  virtual void Noref() { delete this; }

  bool CheckNoRefs() {
    Grab g(ref_mutex);
    assert(refcount);
    return (--refcount == 0);
  }

private:
  Mutex ref_mutex;
  unsigned refcount;
};

template<class T>
class P {
public:
  P() : it(0) {}
  P(T *_it) : it(_it) {
    if (it) it->Ref();
  }
  ~P() {
    if (it) it->Unref();
  }
  P(const P&_it) : it(_it.it) {
    if (it) it->Ref();
  }
  operator T *() const { return it; }
  T &operator*() const { return *it; }
  T *operator->() const { return it; }
  bool operator!() const { return ! it; }
  T *operator=(T *_it) {
    if (it == _it) return it;
    // This little dance is here just in case Unref happens
    // to call the destructor and the destructor happens
    // to call an exception - we make *sure* that we point
    // at the new value and that the new value is ref'd
    // and ready in case this P<> dies before we call Unref().
    T *old_it = it;
    it = _it;
    if (it) it->Ref();
    if (old_it) old_it->Unref();
    return it;
  }
  T *operator=(const P&_it) {
    if (it == _it.it) return it;
    // This little dance is here just in case Unref happens
    // to call the destructor and the destructor happens
    // to call an exception - we make *sure* that we point
    // at the new value and that the new value is ref'd
    // and ready in case this P<> dies before we call Unref().
    T *old_it = it;
    it = _it.it;
    if (it) it->Ref();
    if (old_it) old_it->Unref();
    return it;
  }
protected:
  T *it;
};

#endif // ACAP_REFCOUNT_H
